from functools import singledispatchmethod
from typing import Any

import numpy as np
from numpy.random import permutation

from classiq.interface.generator.expressions.expression import Expression
from classiq.interface.generator.functions.builtins.internal_operators import (
    CONTROL_OPERATOR_NAME,
    INVERT_OPERATOR_NAME,
    WITHIN_APPLY_NAME,
)
from classiq.interface.model.allocate import Allocate
from classiq.interface.model.bind_operation import BindOperation
from classiq.interface.model.classical_if import ClassicalIf
from classiq.interface.model.control import Control
from classiq.interface.model.inplace_binary_operation import InplaceBinaryOperation
from classiq.interface.model.invert import Invert
from classiq.interface.model.model import Model
from classiq.interface.model.native_function_definition import NativeFunctionDefinition
from classiq.interface.model.phase_operation import PhaseOperation
from classiq.interface.model.power import Power
from classiq.interface.model.quantum_expressions.quantum_expression import (
    QuantumAssignmentOperation,
)
from classiq.interface.model.quantum_function_call import QuantumFunctionCall
from classiq.interface.model.quantum_function_declaration import (
    NamedParamsQuantumFunctionDeclaration,
)
from classiq.interface.model.quantum_lambda_function import (
    OperandIdentifier,
    QuantumLambdaFunction,
)
from classiq.interface.model.quantum_statement import QuantumStatement
from classiq.interface.model.repeat import Repeat
from classiq.interface.model.variable_declaration_statement import (
    VariableDeclarationStatement,
)
from classiq.interface.model.within_apply_operation import WithinApply

from classiq.model_expansions.closure import (
    Closure,
    FunctionClosure,
    GenerativeClosure,
    GenerativeFunctionClosure,
)
from classiq.model_expansions.generative_functions import emit_generative_statements
from classiq.model_expansions.interpreters.base_interpreter import BaseInterpreter
from classiq.model_expansions.quantum_operations import (
    BindEmitter,
    ClassicalIfEmitter,
    QuantumFunctionCallEmitter,
    RepeatEmitter,
    VariableDeclarationStatementEmitter,
)
from classiq.model_expansions.quantum_operations.shallow_emitter import ShallowEmitter
from classiq.model_expansions.scope import Evaluated, Scope
from classiq.model_expansions.scope_initialization import (
    add_functions_to_scope,
    add_generative_functions_to_scope,
)
from classiq.qmod.builtins.functions import permute
from classiq.qmod.model_state_container import QMODULE, ModelStateContainer
from classiq.qmod.quantum_function import GenerativeQFunc


class GenerativeInterpreter(BaseInterpreter):
    def __init__(
        self,
        model: Model,
        generative_functions: list[GenerativeQFunc],
    ) -> None:
        super().__init__(model)
        add_generative_functions_to_scope(generative_functions, self._top_level_scope)

    def evaluate_lambda(self, function: QuantumLambdaFunction) -> Evaluated:
        renamed_params = [
            param.rename(function.pos_rename_params[idx])
            for idx, param in enumerate(function.func_decl.positional_arg_declarations)
        ]
        func_decl = NamedParamsQuantumFunctionDeclaration(
            name=self._counted_name_allocator.allocate(
                function.func_decl.name or "<lambda>"
            ),
            positional_arg_declarations=renamed_params,
        )

        closure_class: type[FunctionClosure]
        extra_args: dict[str, Any]
        if function.is_generative():
            closure_class = GenerativeFunctionClosure
            extra_args = {
                "generative_blocks": {
                    "body": GenerativeQFunc(function.py_callable, func_decl),
                }
            }
        else:
            closure_class = FunctionClosure
            extra_args = {}

        return Evaluated(
            value=closure_class.create(
                name=func_decl.name,
                positional_arg_declarations=func_decl.positional_arg_declarations,
                body=function.body,
                scope=Scope(parent=self._builder.current_scope),
                is_lambda=True,
                **extra_args,
            ),
            defining_function=self._builder.current_function,
        )

    @singledispatchmethod
    def emit(self, statement: QuantumStatement) -> None:  # type:ignore[override]
        raise NotImplementedError(f"Cannot emit {statement!r}")

    @emit.register
    def _emit_quantum_function_call(self, call: QuantumFunctionCall) -> None:
        self.emit_quantum_function_call(call)

    def emit_quantum_function_call(self, call: QuantumFunctionCall) -> None:
        QuantumFunctionCallEmitter(self).emit(call)

    @emit.register
    def emit_allocate(self, allocate: Allocate) -> None:
        ShallowEmitter(self, "allocate", components=["size", "target"]).emit(allocate)

    @emit.register
    def emit_bind(self, bind: BindOperation) -> None:
        BindEmitter(self).emit(bind)

    @emit.register
    def emit_quantum_assignment_operation(self, op: QuantumAssignmentOperation) -> None:
        ShallowEmitter(
            self, "assignment_operation", components=["expression", "result_var"]
        ).emit(op)

    @emit.register
    def emit_inplace_binary_operation(self, op: InplaceBinaryOperation) -> None:
        ShallowEmitter(
            self, "inplace_binary_operation", components=["target", "value"]
        ).emit(op)

    @emit.register
    def emit_variable_declaration(
        self, variable_declaration: VariableDeclarationStatement
    ) -> None:
        VariableDeclarationStatementEmitter(self).emit(variable_declaration)

    @emit.register
    def emit_classical_if(self, classical_if: ClassicalIf) -> None:
        ClassicalIfEmitter(self).emit(classical_if)

    @emit.register
    def emit_within_apply(self, within_apply: WithinApply) -> None:
        ShallowEmitter(
            self,
            WITHIN_APPLY_NAME,
            components=["within", "apply", "compute", "action"],
        ).emit(within_apply)

    @emit.register
    def emit_invert(self, invert: Invert) -> None:
        ShallowEmitter(self, INVERT_OPERATOR_NAME, components=["body"]).emit(invert)

    @emit.register
    def emit_repeat(self, repeat: Repeat) -> None:
        RepeatEmitter(self).emit(repeat)

    @emit.register
    def emit_control(self, control: Control) -> None:
        ShallowEmitter(
            self,
            CONTROL_OPERATOR_NAME,
            components=["expression", "body", "else_block"],
        ).emit(control)

    @emit.register
    def emit_power(self, power: Power) -> None:
        ShallowEmitter(self, CONTROL_OPERATOR_NAME, components=["power", "body"]).emit(
            power
        )

    @emit.register
    def emit_phase(self, phase: PhaseOperation) -> None:
        ShallowEmitter(self, "phase", components=["expression", "theta"]).emit(phase)

    def _expand_body(self, operation: Closure) -> None:
        if isinstance(operation, FunctionClosure) and operation.name == "permute":
            # special expansion since permute is generative
            self._expand_permute()
        elif isinstance(operation, GenerativeClosure):
            args = [
                self.evaluate(param.name)
                for param in operation.positional_arg_declarations
            ]
            emit_generative_statements(self, operation, args)
        else:
            super()._expand_body(operation)

    def _expand_permute(self) -> None:
        functions = self.evaluate("functions").as_type(list)
        functions_permutation = permutation(np.array(range(len(functions))))
        calls: list[QuantumFunctionCall] = []
        for function_index in functions_permutation:
            permute_call = QuantumFunctionCall(
                function=OperandIdentifier(
                    name="functions", index=Expression(expr=f"{function_index}")
                )
            )
            permute_call.set_func_decl(permute.func_decl)
            calls.append(permute_call)
        self._expand_block(calls, "body")

    def update_generative_functions(
        self, generative_functions: dict[str, GenerativeQFunc]
    ) -> None:
        add_generative_functions_to_scope(
            list(generative_functions.values()), self._top_level_scope
        )
        for name, gen_func in generative_functions.items():
            if gen_func.compilation_metadata is not None:
                self._functions_compilation_metadata[name] = (
                    gen_func.compilation_metadata
                )

    def update_declarative_functions(
        self,
        functions: dict[str, NativeFunctionDefinition],
        qmodule: ModelStateContainer,
    ) -> None:
        add_functions_to_scope(list(functions.values()), self._top_level_scope)
        for dec_func_name in functions:
            if dec_func_name in qmodule.functions_compilation_metadata:
                self._functions_compilation_metadata[dec_func_name] = (
                    qmodule.functions_compilation_metadata[dec_func_name]
                )

    def add_purely_declarative_function(self, function: FunctionClosure) -> None:
        functions_to_add = [function.name] + QMODULE.function_dependencies[
            function.name
        ]
        for func in functions_to_add:
            if func not in self._expanded_functions and func in QMODULE.native_defs:
                self._expanded_functions[func] = QMODULE.native_defs[func]
                if func in QMODULE.functions_compilation_metadata:
                    self._expanded_functions_compilation_metadata[func] = (
                        QMODULE.functions_compilation_metadata[func]
                    )
